
//
// This source file is part of appleseed.
// Visit http://appleseedhq.net/ for additional information and resources.
//
// This software is released under the MIT license.
//
// Copyright (c) 2010-2013 Francois Beaune, Jupiter Jazz Limited
// Copyright (c) 2014 Francois Beaune, The appleseedhq Organization
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// Interface header.
#include "scenepickinghandler.h"

// appleseed.studio headers.
#include "mainwindow/project/projectexplorer.h"
#include "utility/mousecoordinatestracker.h"

// appleseed.renderer headers.
#include "renderer/api/bsdf.h"
#include "renderer/api/camera.h"
#include "renderer/api/edf.h"
#include "renderer/api/entity.h"
#include "renderer/api/frame.h"
#include "renderer/api/log.h"
#include "renderer/api/material.h"
#include "renderer/api/object.h"
#include "renderer/api/project.h"
#include "renderer/api/rendering.h"
#include "renderer/api/scene.h"
#include "renderer/api/surfaceshader.h"

// appleseed.foundation headers.
#include "foundation/image/color.h"
#include "foundation/image/image.h"
#include "foundation/math/vector.h"

// Qt headers.
#include <QComboBox>
#include <QEvent>
#include <QLabel>
#include <QMouseEvent>
#include <QString>
#include <Qt>
#include <QWidget>

// Standard headers.
#include <cassert>
#include <cstddef>
#include <sstream>

using namespace foundation;
using namespace renderer;
using namespace std;

namespace appleseed {
namespace studio {

ScenePickingHandler::ScenePickingHandler(
    QWidget*                        widget,
    QComboBox*                      picking_mode_combo,
    QLabel*                         r_label,
    QLabel*                         g_label,
    QLabel*                         b_label,
    QLabel*                         a_label,
    const MouseCoordinatesTracker&  mouse_tracker,
    const ProjectExplorer&          project_explorer,
    const Project&                  project)
  : m_widget(widget)
  , m_picking_mode_combo(picking_mode_combo)
  , m_r_label(r_label)
  , m_g_label(g_label)
  , m_b_label(b_label)
  , m_a_label(a_label)
  , m_mouse_tracker(mouse_tracker)
  , m_project_explorer(project_explorer)
  , m_project(project)
  , m_enabled(true)
{
    m_widget->installEventFilter(this);

    m_picking_mode_combo->clear();
    m_picking_mode_combo->addItem("Assembly", "assembly");
    m_picking_mode_combo->addItem("Assembly Instance", "assembly_instance");
    m_picking_mode_combo->addItem("Object", "object");
    m_picking_mode_combo->addItem("Object Instance", "object_instance");
    m_picking_mode_combo->addItem("Material", "material");
    m_picking_mode_combo->addItem("Surface Shader", "surface_shader");
    m_picking_mode_combo->addItem("BSDF", "bsdf");
    m_picking_mode_combo->addItem("EDF", "edf");
    m_picking_mode_combo->setCurrentIndex(3);
}

ScenePickingHandler::~ScenePickingHandler()
{
    m_widget->removeEventFilter(this);
}

void ScenePickingHandler::set_enabled(const bool enabled)
{
    m_enabled = enabled;
}

bool ScenePickingHandler::eventFilter(QObject* object, QEvent* event)
{
    if (!m_enabled)
        return QObject::eventFilter(object, event);

    switch (event->type())
    {
      case QEvent::MouseButtonPress:
        {
            const QMouseEvent* mouse_event = static_cast<QMouseEvent*>(event);
            if (mouse_event->button() == Qt::LeftButton &&
                !(mouse_event->modifiers() & (Qt::AltModifier | Qt::ShiftModifier | Qt::ControlModifier)))
            {
                pick(mouse_event->pos());
                return true;
            }
        }
        break;

      case QEvent::MouseMove:
        {
            const QMouseEvent* mouse_event = static_cast<QMouseEvent*>(event);
            set_rgba_label(mouse_event->pos());
        }
        break;

      case QEvent::Leave:
        m_r_label->clear();
        m_g_label->clear();
        m_b_label->clear();
        m_a_label->clear();
        break;
    }

    return QObject::eventFilter(object, event);
}

namespace
{
    const char* get_primitive_type_name(const ShadingPoint::PrimitiveType primitive_type)
    {
        switch (primitive_type)
        {
          case ShadingPoint::PrimitiveNone: return "none";
          case ShadingPoint::PrimitiveTriangle: return "triangle";
          case ShadingPoint::PrimitiveCurve: return "curve";
          default: return "unknown";
        }
    }

    string print_entity(const char* label, const Entity* entity)
    {
        stringstream sstr;
        sstr << label;

        if (entity)
            sstr << "\"" << entity->get_name() << "\" (#" << entity->get_uid() << ")";
        else sstr << "n/a";

        return sstr.str();
    }

    const Entity* get_picked_entity(
        const ScenePicker::PickingResult&   picking_result,
        const QString&                      picking_mode)
    {
        if (picking_mode == "assembly")
            return picking_result.m_assembly;
        else if (picking_mode == "assembly_instance")
            return picking_result.m_assembly_instance;
        else if (picking_mode == "object")
            return picking_result.m_object;
        else if (picking_mode == "object_instance")
            return picking_result.m_object_instance;
        else if (picking_mode == "material")
            return picking_result.m_material;
        else if (picking_mode == "surface_shader")
            return picking_result.m_surface_shader;
        else if (picking_mode == "bsdf")
            return picking_result.m_bsdf;
        else if (picking_mode == "edf")
            return picking_result.m_edf;
        else
        {
            assert(!"Invalid picking mode.");
            return 0;
        }
    }
}

void ScenePickingHandler::pick(const QPoint& point)
{
    if (!m_project.has_trace_context())
    {
        RENDERER_LOG_INFO("the scene must be rendering or must have been rendered at least once for picking to be available.");
        return;
    }

    const Vector2i pix = m_mouse_tracker.widget_to_pixel(point);
    const Vector2d ndc = m_mouse_tracker.widget_to_ndc(point);

    const ScenePicker scene_picker(m_project.get_trace_context());
    const ScenePicker::PickingResult result = scene_picker.pick(ndc);

    stringstream sstr;

    sstr << "picking details:" << endl;
    sstr << "  pixel coords     " << pix.x << ", " << pix.y << endl;
    sstr << "  ndc coords       " << ndc.x << ", " << ndc.y << endl;
    sstr << "  depth            " << result.m_distance << endl;
    sstr << "  primitive type   " << get_primitive_type_name(result.m_primitive_type) << endl;

    sstr << print_entity("  camera           ", result.m_camera) << endl;
    sstr << print_entity("  assembly inst.   ", result.m_assembly_instance) << endl;
    sstr << print_entity("  assembly         ", result.m_assembly) << endl;
    sstr << print_entity("  object inst.     ", result.m_object_instance) << endl;
    sstr << print_entity("  object           ", result.m_object) << endl;
    sstr << print_entity("  material         ", result.m_material) << endl;
    sstr << print_entity("  surface shader   ", result.m_surface_shader) << endl;
    sstr << print_entity("  bsdf             ", result.m_bsdf) << endl;
    sstr << print_entity("  edf              ", result.m_edf);

    RENDERER_LOG_INFO("%s", sstr.str().c_str());

    const QString picking_mode =
        m_picking_mode_combo->itemData(m_picking_mode_combo->currentIndex()).value<QString>();

    const Entity* picked_entity = get_picked_entity(result, picking_mode);

    if (picked_entity)
        m_project_explorer.highlight_entity(picked_entity->get_uid());
    else m_project_explorer.clear_highlighting();

    emit signal_entity_picked();
}

void ScenePickingHandler::set_rgba_label(const QPoint& point) const
{
    const Vector2i pixel = m_mouse_tracker.widget_to_pixel(point);

    Color4f linear_rgba;
    m_project.get_frame()->image().get_pixel(
        static_cast<size_t>(pixel.x),
        static_cast<size_t>(pixel.y),
        linear_rgba);

    m_r_label->setText(QString().sprintf("%4.3f", linear_rgba.r));
    m_g_label->setText(QString().sprintf("%4.3f", linear_rgba.g));
    m_b_label->setText(QString().sprintf("%4.3f", linear_rgba.b));
    m_a_label->setText(QString().sprintf("%4.3f", linear_rgba.a));
}

}   // namespace studio
}   // namespace appleseed
